from zodipy import emissivities
from zodipy import parameters
from zodipy._model import InterplanetaryDustModel


PLANCK_2013 = InterplanetaryDustModel(
    components=('cloud', 'band1', 'band2', 'band3', 'ring', 'feature'),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2013
)


PLANCK_2015 = InterplanetaryDustModel(
    components=('cloud', 'band1', 'band2', 'band3'),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2015
)

PLANCK_2018 = InterplanetaryDustModel(
    components=('cloud', 'band1', 'band2', 'band3'),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2018
)

# Single component models used for testing
# ----------------------------------------
_CLOUD = InterplanetaryDustModel(
    components=('cloud',),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2018
)

_BAND1 = InterplanetaryDustModel(
    components=('band1',),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2018
)

_BAND2 = InterplanetaryDustModel(
    components=('band2',),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2018
)

_BAND3 = InterplanetaryDustModel(
    components=('band3',),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2018
)

_BANDS = InterplanetaryDustModel(
    components=('band3',),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2018
)

_RING = InterplanetaryDustModel(
    components=('ring',),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2013
)

_FEATURE = InterplanetaryDustModel(
    components=('feature',),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2013
)

_CIRCUMSOLAR = InterplanetaryDustModel(
    components=('ring','feature',),
    parameters=parameters.K98,
    emissivities=emissivities.PLANCK_2013
)


from typing import Iterable, Dict

from zodipy._emissivity import Emissivity
from zodipy import components


class InterplanetaryDustModel:
    """Class that represents a model of the IPD.
    
    Parameters
    ----------
    components : Iterable
        Iterable containing component labels as strings.
    parameters : dict
        Dictionary containing component parameters.
    emissivities : `zodipy.emissivities.Emissivity`
        Emissivity object.
    """

    def __init__(
        self,
        components: Iterable[str], 
        parameters: Dict[str, Dict[str, float]], 
        emissivities: Emissivity
    ) -> None: 
        """Initilizes a Model object."""
        
        self.components = self._init_components(components, parameters)
        self.emissivities = emissivities

    def _init_components(self, comp_labels: Iterable, parameters: dict) -> dict:
        """Initialize component dictionary."""

        comps = {}
        for label in comp_labels:
            if label.startswith('cloud'):
                comp_type = components.Cloud
            elif label.startswith('band'):
                comp_type = components.Band
            elif label.startswith('ring'):
                comp_type = components.Ring
            elif label.startswith('feature'):
                comp_type = components.Feature

            comps[label] = comp_type(**parameters[label])

        return comps


from dataclasses import dataclass
from typing import Callable

import numpy as np
import scipy


@dataclass
class IntegrationConfig:
    """Config object for the LOS/shell integration in the simulation.
    
    Attributes
    ----------
    R_min : float
        Minimum distance from the observer for which to evaluate the 
        dust density.
    R_max : float 
        Maximum distance from the observer for which to evaluate the 
        dust density.
    n : int
        Number of shells for which to evaluate the dust density.
    integator : func
        Function which performs the integration.
    """

    R_max : float
    n : int
    integrator : Callable[[float, float, int], np.ndarray] = np.trapz

    @property
    def R(self) -> np.ndarray:
        """Linearly spaced grid of distances to shells around the observer."""

        ZERO = np.finfo(float).eps
        return np.expand_dims(np.linspace(ZERO, self.R_max, self.n), axis=1)

    @property
    def dR(self) -> np.ndarray:
        """Distance between grid points in R"""

        return np.diff(self.R)


_CUT_OFF = 6

GOOD = {
    'cloud': IntegrationConfig(R_max=_CUT_OFF, n=500),
    'band1': IntegrationConfig(R_max=_CUT_OFF, n=500),
    'band2': IntegrationConfig(R_max=_CUT_OFF, n=500),
    'band3': IntegrationConfig(R_max=_CUT_OFF, n=500),
    'ring': IntegrationConfig(R_max=2.25, n=200),
    'feature': IntegrationConfig(R_max=1, n=200),
}

DEFAULT = {
    'cloud': IntegrationConfig(R_max=_CUT_OFF, n=250),
    'band1': IntegrationConfig(R_max=_CUT_OFF, n=50),
    'band2': IntegrationConfig(R_max=_CUT_OFF, n=50),
    'band3': IntegrationConfig(R_max=_CUT_OFF, n=50),
    'ring': IntegrationConfig(R_max=2.25, n=50),
    'feature': IntegrationConfig(R_max=1, n=50),
}


from typing import Optional, Union, Iterable
from datetime import datetime

import astropy.units as u
import numpy as np

from zodipy import models
from zodipy import _coordinates as coords
from zodipy import _integration as integ
from zodipy import simulation 

class Zodi:
    """Interface for simulating the Zodiacal emission.
    
    Currently, Zodipy only supports simulation the instantaneous Zodiacal 
    emission. It is possible that TOD simulations will be implemented in 
    future.
    """

    def __init__(
        self, 
        observer : Optional[str] = 'L2',
        observation_times : Optional[Iterable[datetime]] = None,
        model : Optional[models.InterplanetaryDustModel] = models.PLANCK_2018,
        integration_config : Optional[integ.IntegrationConfig] = integ.DEFAULT,
    ) -> None:
        """Initializing the zodi interface.

        The geometric setup of the simulation, the IPD model, and the 
        integration configuration used when integrating up the emission 
        are all configured here in the initialization of the Zodi object.
        
        Parameters
        ----------
        observer : str, optional
            The observer. Defaults to L2.
        observation_times : Iterable, optional
            The times of observation. Must be an iterable containing 
            `datetime` objects. Defaults to a single observeration at the 
            current time.
        model : `zodipy.models.Model`, optional
            The Interplanteary dust model used in the simulation. Defaults 
            to the model used in the Planck 2018 analysis.
        integ : `zodipy.integration.IntegrationConfig`, optional
            Integration config object determining the integration details
            used in the simulation.
        """

        if observation_times is None:
            observation_times = [datetime.now().date]

        observer_locations = [
            coords.get_target_coordinates(observer, time) 
            for time in observation_times
        ]
        earth_locations = [
            coords.get_target_coordinates('earth', time) 
            for time in observation_times
        ]

        self.simulation_strategy = simulation.InstantaneousStrategy(
            model, integration_config, observer_locations, earth_locations
        )

    def get_emission(
        self, 
        nside: int, 
        freq: Union[float, u.Quantity], 
        coord: Optional[str] = 'G',
        return_comps: Optional[bool] = False
    ) -> np.ndarray:
        """Returns the simulated Zodiacal emission in units of MJy/sr.

        Parameters
        ----------
        nside : int
            HEALPIX map resolution parameter.
        freq : float, `astropy.units.Quantity`
            Frequency [GHz] at which to evaluate the IPD model. The 
            frequency should be in units of GHz, unless an `astropy.Quantity`
            object is used, for which it only needs to be compatible with Hz.
        coord : str, optional
            Coordinate system of the output map. Accepted inputs are: 'E', 
            'C', or 'G'. Defaults to 'G' which is the Galactic coordinate 
            system.
        return_comps : bool, optional
            If True, the emission of each component in the model is returned
            separatly in form of an array of shape (`n_comps`, `npix`). 
            Defaults to False.

        Returns
        -------
        emission : `numpy.ndarray`
            Simulated Zodiacal emission in units of MJy/sr.
        """

        if isinstance(freq, u.Quantity):
            freq = freq.to('GHz').value

        emission = self.simulation_strategy.simulate(nside, freq)

        if coord != 'E':
            emission = coords.change_coordinate_system(emission, coord)

        return emission if return_comps else emission.sum(axis=0)


from typing import Union

import astropy.constants as const
import numpy as np


def blackbody_emission(
    T: Union[float, np.ndarray], freq: float
) -> Union[float, np.ndarray]:    
    """Returns the blackbody emission.
    
    Assumes the frequency to be in units of GHz.

    Parameters
    ----------
    freq : float
        Frequency [GHz].
    T : float:
        Temperature of the blackbody [K]. 
    Returns
    -------
    float
        Blackbody emission [W / m^2 Hz sr].
    """

    freq *= 1e9
    term1 = (2*const.h.value*freq**3) / const.c.value**2
    term2 = np.expm1((const.h.value*freq) / (const.k_B.value*T))

    return term1 / term2


def interplanetary_temperature(
    R: Union[float, np.ndarray], T_0: float = 286, delta: float = 0.4668626
) -> Union[float, np.ndarray]:
    """Returns the interplanetary temperature as a function of radius.
    
    Parameters
    ----------
    R : float
        Radial distance from the sun in ecliptic coordinates.
    T_0 : float, optional 
        Temperature of the solar system at R = 1 AU.
    delta : float, optional
        Powerlaw index.

    Returns
    -------
    float : 
        Interplanetary temperature.
    """

    return T_0 * R**-delta


from dataclasses import dataclass
from typing import Dict, Iterable, Tuple

import numpy as np


@dataclass
class Emissivity:
    """Class containing emissivity fits for the Zodiacal components."""

    frequencies : Iterable[float]
    components : Dict[str, Tuple[float]]

    def get_emissivity(self, comp: str, freq: float) -> float:
        """Interpolates in the fitted emissivites.
        
        Parameters
        ----------
        comp : str
            Component label, i.e, 'cloud'.
        freq : float
            Frequency at which to evaluate the Zodiacal emission.

        Returns
        -------
        emissivity : float
            Emissivity scaling factor.
        """

        if not self.frequencies[0] <= freq <= self.frequencies[-1]:
            raise ValueError(f'Frequency is out of range')

        emissivity = np.interp(freq, self.frequencies, self.components[comp])

        return emissivity


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['zodipy']

package_data = \
{'': ['*']}

install_requires = \
['astropy>=4.3.1,<5.0.0',
 'astroquery>=0.4.3,<0.5.0',
 'healpy>=1.15.0,<2.0.0',
 'numpy>=1.21.1,<2.0.0',
 'scipy>=1.7.1,<2.0.0']

setup_kwargs = {
    'name': 'zodipy',
    'version': '0.1.0',
    'description': 'Zodipy is a python tool that simulates the instantaneous Zodiacal emission as seen from an observer.',
    'long_description': "# Zodipy\n\n## Description\nZodipy is a python tool that simulates the instantaneous Zodiacal emission as seen from an observer.\n\n## Usage\nThe following will produce a HEALPIX map at `NSIDE=128` of the simulated\nemission as seen by the Planck satellite today\n```python\nimport zodipy\n\nzodi = zodipy.Zodi(observer='Planck')\nemission = zodi.simulate(nside=256, freq=800)\n```\n\nThe time of observation is defined as following\n```python\nimport datetime\n\ntime = datetime(2010, 1, 1)\nzodi = zodipy.Zodi(observer='L2', observation_time=time)\nemission = zodi.simulate(nside=256, freq=800)\n```\n\nThe simulated emission can be visualized using Healpy and matplotlib\n```python\nimport healpy as hp\nimport matplotlib.pyplot as plt\n\nhp.mollview(\n    emission, \n    norm='hist', \n    unit='W/m^2 Hz sr', \n    title='Zodiacal Emission as seen from L2 (2010-01-01)', \n)\nplt.show()\n```\n![plot](imgs/zodi.png)",
    'author': 'Metin San',
    'author_email': 'metinisan@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/MetinSa/zodipy',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.10',
}


setup(**setup_kwargs)



import datetime
import functools
from math import sin, cos
import warnings

from astroquery.jplhorizons import Horizons
import healpy as hp
import numpy as np


TARGET_ALIASES = {
    'l1' : 'SEMB-L1',
    'l2' : 'SEMB-L2',
    'planck' : 'Planck',
    'wmap' : 'WMAP',
    'earth' : 'Earth-Moon Barycenter',
}


@functools.lru_cache
def get_target_coordinates(target: str, time: datetime.datetime.date) -> np.ndarray:
    """Returns the heliocentric cartesian coordinates of the target.
    
    Parameters
    ----------
    target : str
        Name of the target body.
    time : `datetime.datetime`
        Date and time at which to get the targets coordinates.

    Returns
    -------
    `numpy.ndarray`
        Heliocentric cartesian coordinates of the target.
    """

    if target.lower() in TARGET_ALIASES:
        target = TARGET_ALIASES[target.lower()]
    else:
        warnings.warn(
            'The Zodiacal Emission model is only valid in the immediate '
            'surroundings of the earth'
        )

    stop = time + datetime.timedelta(days=1)
    start = time
    epochs = dict(start=str(start), stop=str(stop), step='1d')
    
    query = Horizons(id=target, id_type='majorbody', location='c@sun', epochs=epochs)
    ephemerides = query.ephemerides()

    R = ephemerides['r'][0]
    lon, lat = ephemerides['EclLon'][0], ephemerides['EclLat'][0]

    x = R * cos(lat) * cos(lon)
    y = R * cos(lat) * sin(lon)
    z = R * sin(lat)
    
    return np.array([[x], [y], [z]])


def change_coordinate_system(
    input_map: np.ndarray, coord_out: str, coord_in: str = 'E'
) -> np.ndarray:
    """Rotates a map from coordinate system to another using healpy.Rotator
    
    Paramters
    ---------
    input_map : `numpy.ndarray`
        Map to rotate.
    coord_out : str
        Coordinate system of the output_map.
    coord_in : str
        Coordinate system of the input_map.
    Returns
    -------
    output_map : `numpy.ndarray`
        Rotated map.
    """

    if coord_in == coord_out:
        return input_map
    
    nside = hp.get_nside(input_map)
    npix = hp.nside2npix(nside)
    
    rotator = hp.Rotator(coord=[coord_out, coord_in])
    new_angles = rotator(hp.pix2ang(nside, np.arange(npix)))
    new_pixels = hp.ang2pix(nside, *new_angles)
    output_map = input_map[..., new_pixels]

    return output_map


from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Iterable

import healpy as hp
import numpy as np

from zodipy._model import InterplanetaryDustModel
from zodipy._integration import IntegrationConfig


@dataclass
class SimulationStrategy(ABC):
    """Class that represents the simulation part of Zodipy.
    
    The simulation strategy is responsible for simulating the Zodiacal
    emission given an initial setup of the problem.

    Parameters
    ----------
    model : `zodipy._model.InterplanetaryDustModel`
        IPD model with initialized componentents and corresponding 
        emissivities.
    integration_config: `zodipy._integ.IntegrationConfig`
        Configuration object that determines how a component is integrated
        along a line of sight.
    observer_locations : Iterable
        Iterable containing the various locations of an observer. One 
        instantaneous simulation is produced per location.
    earth_locations : Iterable
        Iterable containing the various locations of the Earth 
        corresponding to the observer locations.
    """

    model: InterplanetaryDustModel
    integration_config: IntegrationConfig
    observer_locations: Iterable
    earth_locations: Iterable

    @abstractmethod
    def simulate(self, nside: int, freq: float) -> np.ndarray:
        """Simulates the Zodiacal emission, given a nside and frequency.
        
        The emission is returned in units of MJy/sr.

        Parameters
        ----------
        nside : int
            HEALPIX map resolution parameter.
        freq : float
            Frequency [GHz] at which to evaluate the IPD model.

        Returns
        -------
        emission : `np.ndarray`
            Simulated Zodiacal emission.
        """


class InstantaneousStrategy(SimulationStrategy):
    """Simulation strategy that computes the instantaneous emission.
    
    By instantaneous emission, we mean the emission that is seen at one
    instant in time. The emission is averaged over all observations.
    """

    def __init__(
        self, model, integration_config, observer_locations, earth_locations
    ) -> None:
        super().__init__(
            model, integration_config, observer_locations, earth_locations
        )

    def simulate(self, nside: int, freq: float) -> np.ndarray:
        """See base class."""

        X_observer  = self.observer_locations
        X_earth  = self.earth_locations
        X_unit = np.asarray(
            hp.pix2vec(nside, np.arange(npix := hp.nside2npix(nside)))
        )

        emission = np.zeros(
            shape=(
                n_observations := len(X_observer), 
                len(components := self.model.components), 
                npix
            )
        )
        for observation in range(n_observations):
            for comp_idx, (comp_name, comp) in enumerate(components.items()):
                integration_config = self.integration_config[comp_name]
                comp_emission = comp.get_emission(
                    freq, 
                    X_observer[observation], 
                    X_earth[observation], 
                    X_unit, 
                    integration_config.R
                )
                integrated_comp_emission = integration_config.integrator(
                    comp_emission, 
                    integration_config.R, 
                    dx=integration_config.dR, 
                    axis=0
                )

                comp_emissivity = self.model.emissivities.get_emissivity(
                    comp_name, freq
                )
                integrated_comp_emission *= comp_emissivity

                emission[observation, comp_idx] = integrated_comp_emission
        
        return emission.mean(axis=0) * 1e20